--[[
Name: LibStickyFrames-2.0
Revision: $Rev: 73376 $
Author(s): Sorata,Toadkiller,Cladhaire,Tuller,Shadowed,JoshBorke,rophy123,Tekkub
Website: http://www.wowace.com/wiki/LibStickyFrames-2.0
Description: A library to manage snappy dragging & sticking of frames across multiple mods.
Dependencies: none
--]]

local MAJOR = "LibStickyFrames-2.0"
local MINOR = "$Revision: 73376 $"

--[[
A library to manage snappy dragging & sticking of frames across multiple mods.

=The vision:=
Many mods have invisible display areas or otherwise draggable frames that I need to lay out in my UI so that they do not overlap & thus can exist in harmony.  All mods with such frames will one day use this lib to accomplish the following:
* When I start dragging a bar around in Bartender3 or AutoBar, or the raid warnings frames from Deadly Boss Mods, BigWigs or other display frames from Quartz, SCT, etc. then all these frames go green & red and so on & I can see exactly where I can drag to & thus snap to or stick to so I can lay out my interface with or without overlap as I desire.
* Once the sticking code goes in it is even more beautiful since I can just drag a few frames around at that point without destroying my careful layout or getting hosed with a ui scale change...
* Since the dragging interface is standard, I am not uselessly trying ctrl-alt-shift-left&right-button drag just to move some frame around.
* All these hard to access frames suddenly become easily accesible.

=Design goals:=
* A consistent drag experience across implementing mods.
* Snapping & sticking to all registered frames + select Blizzard frames.
* Visibility & Labeling of all frames of interest while dragging (except for the Blizzard frames?)
* Private Clusters
** Private Clusters stick only to themselves.
** Private Clusters have a supplied or calculated clusterFrame which participates in general sticking.
* Grouped Frames
** Groups of frames that are stuck together can be dragged by dragging any member of the group
* Consistent interface for all these operations, including modifier keys for sticking or not snapping or unsticking etc.

=Planned Features:=
* Make frame snapping work better

=Origin of the Species=
This is based on code from:
[http://www.wowinterface.com/downloads/info4674-StickyFrames.html StickyFrames] by Cladhaire.
[http://www.wowwiki.com/LegoBlock LegoBlock] by Tekkub & JoshBorke.
FlyPaper by Tuller.
All mushed together in a terrible transporter accident producing the thing otherwise known as lib by Toadkiller enhanced by Sorata.
--]]

local lib, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not lib then return end -- no upgrade needed

local _G = _G

-- CallbackHandler
lib.events = lib.events or _G.LibStub("CallbackHandler-1.0"):New(lib)
-- Current Sticky Group
lib.group = lib.group or false
-- Embeded Frames
lib.registered = lib.registered or {}
lib.registered[_G.MainMenuBarArtFrame] = true
lib.registered[_G.CharacterMicroButton] = true
lib.registered[_G.ChatFrameMenuButton] = true
--lib.registered[_G.MainMenuBar] = true
lib.registered[_G.ChatFrame1] = true
-- Frame Insets
lib.insets = lib.insets or {}
lib.insets[_G.CharacterMicroButton] = { 3, 23, -181, 1 }
lib.insets[_G.ChatFrameMenuButton] = { 3, 3, 3, -91 }
lib.insets[_G.MainMenuBar] = { 7, 0, 40, 0 }
lib.insets[_G.ChatFrame1] = { -2, -4, -2, -6, }
-- Enabled Frames
lib.enabled = lib.enabled or {}
-- Hidden Frames
lib.hidden = lib.hidden or {}
-- Frame Anchors
lib.anchors = lib.anchors or {}
-- Frame Overlays
lib.overlays = lib.overlays or {}
lib.overlayCache = lib.overlayCache or {}
-- Frame Groups
lib.groups = lib.groups or {}
-- Frame Texts
lib.texts = lib.texts or {}
-- Table to temporarily hold any existing OnUpdate scripts.
lib.scripts = lib.scripts or {}

-- Snap range in pixels.  not saved, but mods can provide an interface to change it
lib.snapRangeDefault = 5
lib.snapRange = lib.snapRange or lib.snapRangeDefault

-- Key to hold down for no snapping.  Set to IsShiftKeyDown, IsControlKeyDown or IsAltKeyDown functions.
lib.IsSnapDisableKeyDown = lib.IsSnapDisableKeyDown or IsShiftKeyDown

-- Default color values to use.
lib.colorHidden = lib.colorHidden or { 1, 0, 0, 0.5 }
lib.colorEnabled = lib.colorEnabled or { 0, 1, 0, 0.5 }
lib.colorDisabled = lib.colorDisabled or { 0, 0, 1, 0.5 }

-- Default border color values for hovering over the frame
lib.colorBorderEnabled = lib.colorBorderEnabled or { 0.5, 0.5, 0, 1 }
lib.colorBorderDisabled = lib.colorBorderDisabled or { 0, 0, 0, 0 }

--[[
Internal functions
--]]

local RetrieveOverlay
local StoreOverlay
local GetFramePoints
local IsOverlapping
local GetInsetOffset
local CanSnapFrame
local IsDependentOnFrame
local CanAttach
local MovingUpdate
local UpdateFrameGroup

function RetrieveOverlay(self)
	local overlay = next(lib.overlayCache)
	if overlay then
		lib.overlayCache[overlay] = nil
	else
		overlay = CreateFrame("Button", nil, self)
		overlay:EnableMouse(true)
		overlay:RegisterForDrag("LeftButton")
		overlay:RegisterForClicks("RightButtonDown", "LeftButtonUp")
		overlay:SetToplevel(true)
		overlay:SetBackdrop({ bgFile = [[Interface\Tooltips\UI-Tooltip-Background]], tile = true, tileSize = 16, edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]], edgeSize = 16, insets = {}, })
		overlay:SetAllPoints(self)
		overlay:Hide()
		local anchor = overlay:CreateTexture(nil, "ARTWORK")
		anchor:SetTexture([[Interface\Tooltips\UI-Tooltip-Background]])
		anchor:SetVertexColor(1, 0, 0, 0.5)
		anchor:ClearAllPoints()
		anchor:Hide()
		overlay.anchor = anchor
		local text = overlay:CreateFontString(nil, "OVERLAY")
		text:SetFontObject(GameFontNormal)
		text:ClearAllPoints()
		text:SetPoint("CENTER", overlay, "CENTER", 0,0)
		overlay.text = text
	end
	lib.overlays[self] = overlay
	overlay:SetParent(self)
	overlay:SetFrameStrata("DIALOG")
	lib:SetFrameEnabled(self, lib:IsFrameEnabled(self))
	lib:SetFrameInsets(self, lib:GetFrameInsets(self))
	lib:SetFrameAnchor(self, lib:GetFrameAnchor(self))
	lib:SetFrameText(self, lib:GetFrameText(self))
	return overlay
end

function StoreOverlay(self)
	local overlay = lib.overlays[self]
	if overlay then
		overlay:ClearAllPoints()
		overlay:SetParent(nil)
		overlay:SetScript("OnEnter", nil)
		overlay:SetScript("OnLeave", nil)
		overlay:SetScript("OnDragStart", nil)
		overlay:SetScript("OnDragStop", nil)
		overlay:Hide()
		lib.overlays[self] = nil
		lib.overlayCache[overlay] = true
	end
end

function GetFramePoints(frame)
	local l, t, r, b = lib:GetFrameInsets(frame)
	return frame:GetLeft() + l, frame:GetTop() - t, frame:GetRight() - r, frame:GetBottom() + b
end

--[[
Determines the overlap between two frames.  Returns true if the frames
overlap anywhere, or false if they do not. Insets are automatically taken into
account.
--]]
function IsOverlapping(frameA, frameB)
	local lA, tA, rA, bA = GetFramePoints(frameA)
	local lB, tB, rB, bB = GetFramePoints(frameB)
	local sA, sB = frameA:GetEffectiveScale(), frameB:GetEffectiveScale()
	return  ((lA * sA) < (rB * sB))
		and ((lB * sB) < (rA * sA))
		and ((bA * sA) < (tB * sB))
		and ((bB * sB) < (tA * sA))
end

--[[
This returns the x and y offsets based on the frame insets and the anchor points

frameA:	The frame to be anchored
pointA:	The point on frameA to be attached
frameB:	The frame being anchored to
pointB:	The point on frameB to be attached to
--]]
function GetInsetOffset(frameA, pointA, frameB, pointB)
	pointA, pointB = pointA:lower(), (pointB or pointA):lower()
	local sA, sB = frameA:GetEffectiveScale(), frameB:GetEffectiveScale()

	local lA, tA, rA, bA = lib:GetFrameInsets(frameA)
	lA, tA, rA, bA = lA * sA, tA * sA, rA * sA, bA * sA
	if pointA ~= "left" and pointA ~= "topleft" and pointA ~= "bottomleft" then
		lA = 0
	end
	if pointA ~= "right" and pointA ~= "topright" and pointA ~= "bottomright" then
		rA = 0
	end
	if pointA ~= "top" and pointA ~= "topleft" and pointA ~= "topright" then
		tA = 0
	end
	if pointA ~= "bottom" and pointA ~= "bottomleft" and pointA ~= "bottomright" then
		bA = 0
	end

	local lB, tB, rB, bB = lib:GetFrameInsets(frameB)
	lB, tB, rB, bB = lB * sB, tB * sB, rB * sB, bB * sB
	if pointB ~= "left" and pointB ~= "topleft" and pointB ~= "bottomleft" then
		lB = 0
	end
	if pointB ~= "right" and pointB ~= "topright" and pointB ~= "bottomright" then
		rB = 0
	end
	if pointB ~= "top" and pointB ~= "topleft" and pointB ~= "topright" then
		tB = 0
	end
	if pointB ~= "bottom" and pointB ~= "bottomleft" and pointB ~= "bottomright" then
		bB = 0
	end

	local l, t, r, b
	if (pointA == "left" or pointA == "topleft" or pointA == "bottomleft") and (pointB == "left" or pointB == "topleft" or pointB == "bottomleft") then
		l = lA - lB
	else
		l = lA + lB
	end
	if (pointA == "top" or pointA == "topleft" or pointA == "topright") and (pointB == "top" or pointB == "topleft" or pointB == "topright") then
		t = tA - tB
	else
		t = tA + tB
	end
	if (pointA == "right" or pointA == "topright" or pointA == "bottomright") and (pointB == "right" or pointB == "topright" or pointB == "bottomright") then
		r = rA - rB
	else
		r = rA + rB
	end
	if (pointA == "bottom" or pointA == "bottomleft" or pointA == "bottomright") and (pointB == "bottom" or pointB == "bottomleft" or pointB == "bottomright") then
		b = bA - bB
	else
		b = bA + bB
	end

	local x = (l + r) / sA
	if (pointA == "left" or pointA == "topleft" or pointA == "bottomleft") then
		x = x * -1
	end
	local y = (b + t) / sA
	if (pointA == "bottom" or pointA == "bottomleft" or pointA == "bottomright") then
		y = y * -1
	end
	
	return x, y
end

--[[
This is called when finding an overlap between two sticky frames.  If frameA is near
a sticky point of frameB, then it will snap to that point and return true.  If there
is no sticky point collision, it returns nil so we can test other frames for
stickyness.
--]]
function CanSnapFrame(frameA, frameB)
	local sA, sB = frameA:GetEffectiveScale(), frameB:GetEffectiveScale()
	local xA, yA = frameA:GetCenter()
	local xB, yB = frameB:GetCenter()
	xA, yA, xB, yB = xA * sA, yA * sA, xB * sB, yB * sB

	-- Grab the points of each frame, for easier comparison
	local lA, tA, rA, bA = GetFramePoints(frameA)
	lA, tA, rA, bA = lA * sA, tA * sA, rA * sA, bA * sA
	local lB, tB, rB, bB = GetFramePoints(frameB)
	lB, tB, rB, bB = lB * sB, tB * sB, rB * sB, bB * sB

	local snapRange = lib.snapRange

	local pointA, pointB
	local x, y = 0, 0

	-- If we are within snapRange pixels of the point of a sticky frame then snap to the point.
	if lA > (rB - snapRange) then		-- Left stickyness
		if (yA < (yB + snapRange)) and (yA > (yB - snapRange)) then
			pointA = "left"
			pointB = "right"
		elseif tA > (tB - snapRange) and tA < (tB + snapRange) then
			pointA = "topleft"
			pointB = "topright"
		elseif bA < (bB + snapRange) and bA > (bB - snapRange) then
			pointA = "bottomleft"
			pointB = "bottomright"
		else
			pointA = "left"
			pointB = "right"
			if (yA > (yB + snapRange)) or (yA < (yB - snapRange)) then
				y = (yA - yB) / sA
			end
		end
	elseif rA < (lB + snapRange) then	-- Right stickyness
		if (yA < (yB + snapRange)) and (yA > (yB - snapRange)) then
			pointA = "right"
			pointB = "left"
		elseif tA > (tB - snapRange) and tA < (tB + snapRange) then
			pointA = "topright"
			pointB = "topleft"
		elseif bA < (bB + snapRange) and bA > (bB - snapRange) then
			pointA = "bottomright"
			pointB = "bottomleft"
		else
			pointA = "right"
			pointB = "left"
			if (yA > (yB + snapRange)) or (yA < (yB - snapRange)) then
				y = (yA - yB) / sA
			end
		end
	elseif bA > (tB - snapRange) then	-- Bottom stickyness
		if (xA < (xB + snapRange)) and (xA > (xB - snapRange)) then
			pointA = "bottom"
			pointB = "top"
		elseif lA > (lB - snapRange) and lA < (lB + snapRange) then
			pointA = "bottomleft"
			pointB = "topleft"
		elseif rA < (rB + snapRange) and rA > (rB - snapRange) then
			pointA = "bottomright"
			pointB = "topright"
		else
			pointA = "bottom"
			pointB = "top"
			if (xA > (xB + snapRange)) or (xA < (xB - snapRange)) then
				x = (xA - xB) / sA
			end
		end
	elseif tA < (bB + snapRange) then	-- Top stickyness
		if (xA < (xB + snapRange)) and (xA > (xB - snapRange)) then
			pointA = "top"
			pointB = "bottom"
		elseif lA > (lB - snapRange) and lA < (lB + snapRange) then
			pointA = "topleft"
			pointB = "bottomleft"
		elseif rA < (rB + snapRange) and rA > (rB - snapRange) then
			pointA = "topright"
			pointB = "bottomright"
		else
			pointA = "top"
			pointB = "bottom"
			if (xA > (xB + snapRange)) or (xA < (xB - snapRange)) then
				x = (xA - xB) / sA
			end
		end
	end

	-- Snap to the found point
	if pointA then
		local xOffset, yOffset = GetInsetOffset(frameA, pointA, frameB, pointB)
		frameA:ClearAllPoints()
		frameA:SetPoint(pointA, frameB, pointB, x + xOffset, y + yOffset)
		lib:SetFrameAnchor(frameA, pointA)

		-- Remember Sticking stuff
		lib.point = pointA
		lib.frameTo = frameB
		lib.pointTo = pointB
		lib.x = x
		lib.y = y
		lib.events:Fire("OnStickToFrame", frameA, pointA, frameB, pointB, x, y)
		return true
	end
end

--[[
FLYPAPER
Handles sticking frames to eachother using relative positioning
credit to Tuller for this
--]]
function IsDependentOnFrame(frameA, frameB)
	if not frameA or not frameB then
		return
	elseif frameA == frameB then
		return true
	end

	local points = frameA:GetNumPoints()
	for i = 1, points do
		local _, frameC = frameA:GetPoint(i)
		if IsDependentOnFrame(frameC, frameB) then
			return true
		end
	end
end

--[[
returns true if its actually possible to attach the two frames without error
--]]
function CanAttach(frameA, frameB)
	if frameA and frameB and frameA:GetWidth() ~= 0 and frameA:GetHeight() ~= 0 and frameB:GetWidth() ~= 0 and frameB:GetHeight() ~= 0 and not IsDependentOnFrame(frameB, frameA) then
		return true
	end
end

--[[
Returns an anonymous OnUpdate function for the frame in question.
--]]
function MovingUpdate()
	local frame = lib.frame
	local x, y = GetCursorPosition()
	local s = frame:GetEffectiveScale()

	x, y = (x / s) + lib.cursorX, (y / s) + lib.cursorY

	frame:ClearAllPoints()
	frame:SetPoint("CENTER", UIParent, "BOTTOMLEFT", x, y)

	if not lib.IsSnapDisableKeyDown() then
		local group = lib:GetFrameGroup(frame)
		for frameB in pairs(lib.registered) do
			if frame ~= frameB and frameB:IsVisible() and (lib:InFrameGroup(frameB, group) or group == frameB or (type(group) == "table" and group[frameB])) and CanAttach(frame, frameB) and IsOverlapping(frame, frameB) and CanSnapFrame(frame, frameB) then
				return
			end
		end
	end

	lib.point = nil
	lib.frameTo = nil
	lib.pointTo = nil
	lib.x = nil
	lib.y = nil
	lib:SetFrameAnchor(frame, nil)
end

do
	local function onShow(self)
		lib:UpdateFrameColor(self:GetParent())
	end
	local onEnter = onShow
	local onLeave = onShow
	local function onClick(self, button)
		lib.events:Fire("OnClick", self:GetParent(), button)
	end
	local function onDragStart(self)
		lib:StartFrameMoving(self:GetParent())
	end
	local function onDragStop(self)
		lib:StopFrameMoving(self:GetParent())
	end
	--[[
	UpdateFrameGroup(frame) - 
	--]]
	function UpdateFrameGroup(frame)
		local overlay
		local group = lib.group
		local inGroup = lib:InFrameGroup(frame, group)
		if inGroup or group == frame or (type(group) == "table" and group[frame]) then
			overlay = lib:GetFrameOverlay(frame)

			overlay:SetScript("OnShow", onShow)
			overlay:SetScript("OnClick", onClick)
			if inGroup and lib:IsFrameEnabled(frame) then
				overlay:SetScript("OnEnter", onEnter)
				overlay:SetScript("OnLeave", onLeave)
				overlay:SetScript("OnDragStart", onDragStart)
				overlay:SetScript("OnDragStop", onDragStop)
			else
				overlay:SetScript("OnEnter", nil)
				overlay:SetScript("OnLeave", nil)
				overlay:SetScript("OnDragStart", nil)
				overlay:SetScript("OnDragStop", nil)
			end
	
			overlay:Show()
			if not frame:IsShown() then
				lib:SetFrameHidden(frame, true)
				frame:Show()
			end
			return true
		else
			StoreOverlay(frame)
			if lib:IsFrameMoving(frame) then
				lib:StopFrameMoving(frame)
			end
			if lib:IsFrameHidden(frame) then
				frame:Hide()
				lib:SetFrameHidden(frame, false)
			end
		end
	end
end

--[[
Library functions
--]]

function lib:RegisterFrame(self)
	lib.registered[self] = true
	UpdateFrameGroup(self)
end

function lib:IsRegisteredFrame(self)
	return lib.registered[self]
end

function lib:UnregisterFrame(self)
	lib.registered[self] = nil
	lib.insets[self] = nil
	lib.enabled[self] = nil
	lib.anchors[self] = nil
	lib.groups[self] = nil
	lib.texts[self] = nil
	UpdateFrameGroup(self)
end

--[[
lib:SetGroup(group) - Set Sticky Mode to be on or off.
group: anything. true will make all frames sticky while anything else will only
make frames with the same group sticky.
--]]
function lib:SetGroup(group)
	group = group and group or nil
	if (lib.group ~= group) then
		lib.group = group
		for frame in pairs(lib.registered) do
			UpdateFrameGroup(frame)
		end
		lib.events:Fire("OnSetGroup", group)
	end
end

--[[
lib:GetGroup() - returns the current Sticky Group
--]]
function lib:GetGroup()
	return lib.group
end

--[[
lib:AnchorFrame(frame) - This can be called in conjunction with self:StopStickyMoving() to anchor the
frame right back to the parent, so you can manipulate its children as a group
(This is useful in WatchDog)
--]]
function lib:AnchorFrame(frame)
	local xA,yA = frame:GetCenter()
	local parent = frame:GetParent() or UIParent
	local xP, yP = parent:GetCenter()
	local sA, sP = frame:GetEffectiveScale(), parent:GetEffectiveScale()

	xP,yP = (xP * sP) / sA, (yP * sP) / sA

	local xo,yo = (xP - xA) * -1, (yP - yA) * -1

	frame:ClearAllPoints()
	frame:SetPoint("CENTER", parent, "CENTER", xo, yo)
end

--[[
lib:AttachFrameTo(frameA, pointA, frameB, pointB, x, y) - Convenience method to Stick

frameA:	The frame that was being dragged
pointA:	The point on frame that is sticking
frameB:	The frame to stick to
pointB:	The point on frameTo to stick to
x: 		The x offset (the insets offset is automatically added)
y:		the y offset (the insets offset is automatically added)
--]]
function lib:AttachFrameTo(frameA, pointA, frameB, pointB, x, y)
	if not frameA or not frameB or IsDependentOnFrame(frameB, frameA) then
		return
	end
	pointA, pointB = pointA:lower(), (pointB or pointA):lower()

	local xOffset, yOffset = GetInsetOffset(frameA, pointA, frameB, pointB, x, y)
	frameA:ClearAllPoints()
	frameA:SetPoint(pointA, frameB, pointB, (x or 0) + xOffset, (y or 0) + yOffset)
	lib:SetFrameAnchor(frameA, pointA)
end

--[[
self:SetFrameInsets(left, top, right, bottom) - Sets the insets for the frame.

left:		If your frame has a transparent border around the entire frame
			(think backdrops with borders).  This can be used to fine tune the
			edges when you are stickying groups.  Refers to any offset on the
			LEFT edge of the frame being moved.
top:		same
right:		same
bottom:		same
--]]
function lib:SetFrameInsets(self, ...)
	local insets = lib.insets[self] or {}
	for i=1,4 do
		local inset = select(i, ...)
		insets[i] = type(inset) == "number" and inset or 0
	end
	lib.insets[self] = insets

	local overlay = lib:GetFrameOverlay(self, true)
	if overlay then
		overlay:ClearAllPoints()
		overlay:SetPoint("left", self, "left", GetInsetOffset(overlay, "left", self, "left"))
		overlay:SetPoint("top", self, "top", GetInsetOffset(overlay, "top", self, "top"))
		overlay:SetPoint("right", self, "right", GetInsetOffset(overlay, "right", self, "right"))
		overlay:SetPoint("bottom", self, "bottom", GetInsetOffset(overlay, "bottom", self, "bottom"))
	end
end

--[[
left, top, right, bottom = self:GetFrameInsets() - returns the insets for the frame in this order
--]]
function lib:GetFrameInsets(self)
	local insets = lib.insets[self]
	if insets then
		return unpack(insets)
	else
		return 0, 0, 0, 0
	end
end

--[[
self:SetFrameGroup(group) - Sets the group for the frame.
group: anything. true will make all frames sticky while anything else will only
make frames with the same group sticky. If group is a table it must contain the
other frames grouped with it like so: group = { [frame1] = true, [frame2] = true, }
--]]
function lib:SetFrameGroup(self, group)
	group = group and group ~= true and group or nil
	lib.groups[self] = group
	UpdateFrameGroup(self)
end

--[[
group = self:GetFrameGroup() - returns the group for the frame
--]]
function lib:GetFrameGroup(self)
	return lib.groups[self] or true
end

--[[
 - returns true if in the sticky group
--]]
--[[
inGroup = self:InFrameGroup(group) - Returns true if the frames are in the same group or if no group is set on both.
If one or both groups are a table index by frames then both groups will be checked
to see if the frame is present.
--]]
function lib:InFrameGroup(self, groupB)
	local groupA = lib:GetFrameGroup(self)
	if groupA == groupB or groupB == true then
		return true
	end
end

do
	local rotPoints = { top = "right", bottom = "left", left = "top", right = "bottom" }
	local relPoints = { top = "bottom", bottom = "top", left = "right", right = "left", topright = "bottomleft", bottomright = "topleft", bottomleft = "topright", topleft = "bottomright" }
	--[[
	self:SetFrameAnchor(point) - sets the visual indication for the anchor point of the frame.
	--]]
	function lib:SetFrameAnchor(self, point)
		point = point and point or nil
		if point and not relPoints[point] then return end
		lib.anchors[self] = point

		local overlay = lib:GetFrameOverlay(self, true)
		if not overlay then return end
		local anchor = overlay.anchor
		if point then
			local relPoint = relPoints[point]
			local rotPoint = rotPoints[point]
			anchor:ClearAllPoints()
			anchor:SetPoint(point, self, point, 0, 0)
			anchor:SetPoint(relPoint, self, "center", 0, 0)
			if rotPoint then
				local rotRelPoint = rotPoints[relPoint]
				anchor:SetPoint(rotPoint, self, rotPoint, 0, 0)
				anchor:SetPoint(rotRelPoint, self, rotRelPoint, 0, 0)
			end
			anchor:Show()
		else
			anchor:Hide()
		end
	end
end

--[[
point = self:GetFrameAnchor() - returns the anchor point of the frame
--]]
function lib:GetFrameAnchor(self)
	return lib.anchors[self]
end

--[[
overlay = self:GetFrameOverlay() - returns the overlay and creates it if it doesn't exist
--]]
function lib:GetFrameOverlay(self, nocreate)
	if not lib.overlays[self] and not nocreate then
		return RetrieveOverlay(self)
	end
	return lib.overlays[self]
end

function lib:SetFrameText(self, text)
	text = type(text) == "string" and text or nil
	lib.texts[self] = text
	local overlay = lib:GetFrameOverlay(self, true)
	if overlay then
		overlay.text:SetText(text)
	end
end

function lib:GetFrameText(self)
	return lib.texts[self]
end

--[[
self:SetFrameEnabled(state) - Sets whether the frame is enabled or not
state: true or false
--]]
function lib:SetFrameEnabled(self, state)
	state = state and true or nil
	lib.enabled[self] = state
	UpdateFrameGroup(self)
end

--[[
state = self:IsFrameEnabled() - returns the enabled state
--]]
function lib:IsFrameEnabled(self, frame)
	return lib.enabled[self]
end

--[[
self:SetFrameHidden(state) - Set whether the frame should be hidden when lib:SetGroup(nil)
is called and is automatically set if the frame isn't shown when lib:SetGroup(nil).
--]]
function lib:SetFrameHidden(self, state)
	lib.hidden[self] = state and true or nil
	lib:UpdateFrameColor(self)
end

--[[
state = self:IsFrameHidden() - returns the hidden state. Is cleared after lib:SetGroup(nil)
is called.
--]]
function lib:IsFrameHidden(self)
	return lib.hidden[self] and true
end

--[[
self:UpdateFrameColor() - Updates the color of the overlay depending on the state of the frame
--]]
function lib:UpdateFrameColor(self)
	local overlay = lib:GetFrameOverlay(self, true)
	if not overlay then return end
	local color, borderColor
	if lib:IsFrameMoving(self) or GetMouseFocus() == overlay then
		borderColor = lib.colorBorderEnabled
	else
		borderColor = lib.colorBorderDisabled
	end
	local frame = lib.frame
	local group = lib:GetFrameGroup(frame)
	if lib:IsFrameHidden(self) or not self:IsShown() then
		color = lib.colorHidden
	elseif frame then
		if lib:InFrameGroup(self, group) or group == self or (type(group) == "table" and group[self]) then
			color = lib.colorEnabled
		else
			color = lib.colorDisabled
		end
	elseif lib:InFrameGroup(self, lib.group) and lib:IsFrameEnabled(self) then
		color = lib.colorEnabled
	else
		color = lib.colorDisabled
	end
	overlay:SetBackdropColor(unpack(color))
	overlay:SetBackdropBorderColor(unpack(borderColor))
end

--[[
self:StartFrameMoving() - Sets a custom OnUpdate for the frame so it follows
the mouse and snaps to the frames you specify
--]]
function lib:StartFrameMoving(self)
	if lib.frame then
		lib:StopFrameMoving(lib.frame)
	end
	lib.frame = self
	lib:SetFrameAnchor(self, nil)
	for frame in pairs(lib.registered) do
		lib:UpdateFrameColor(frame)
	end

	local x, y = GetCursorPosition()
	local xA, yA = self:GetCenter()
	local sA = self:GetEffectiveScale()
	lib.cursorX = (xA * sA) - x
	lib.cursorY = (yA * sA) - y

	lib.scripts[self] = self:GetScript("OnUpdate")
	self:SetScript("OnUpdate", MovingUpdate)
	lib.events:Fire("OnStartFrameMoving", self)
end

--[[
self:IsFrameMoving() - returns true if the frame is moving.
--]]
function lib:IsFrameMoving(self)
	return lib.frame == self
end

--[[
self:StopFrameMoving() - This stops the OnUpdate, leaving the frame at its last position.  This will
leave it anchored to UIParent.  You can call self:AnchorFrameFrame() to
anchor it back "TOPLEFT" , "TOPLEFT" to the parent.
--]]
function lib:StopFrameMoving(self)
	lib:SetFrameAnchor(self, lib.point)
	-- Perform Sticking if requested
	lib.events:Fire("OnStopFrameMoving", self, lib.point, lib.frameTo, lib.pointTo, lib.x, lib.y)
	lib.frame = nil
	lib.point = nil
	lib.frameTo = nil
	lib.pointTo = nil
	lib.x = nil
	lib.y = nil

	self:SetScript("OnUpdate", lib.scripts[self])
	lib.scripts[self] = nil
	for frame in pairs(lib.registered) do
		lib:UpdateFrameColor(frame)
	end
end